#ifndef _SimpleChecker_h_
#define _SimpleChecker_h_

#include <iostream>
#include <fstream>
#include <iomanip>
#include <stdexcept>
using namespace std;

#ifdef useGMP
#include <gmpxx.h>

typedef 
  mpz_class IntegerType;
#else
#warning this IntegerType may not be suitable for some input file. Consider using GMP
typedef
  long IntegerType;
#endif

class SimpleChecker
{
private:
  IntegerType leftTerm,rightTerm; // left and right terms of a constraint;
  string relOp; // relational operator of a constraint

  // information on a constraint
  string constraintName; // name of a constraint (if any)

  map<string,bool> interpretation; // interpretation of each variable

  IntegerType objectiveValue; // value of the objective function for
			      // this interpretation

  int line;

public:


  /**
   * forget the current interpretation
   */
  void clear()
  {
    interpretation.clear();
  }

  /**
   * @return the number of assigned variables in the current
   * interpretation
   */
  int nbAssignedVariables()
  {
    return interpretation.size();
  }


  IntegerType getObjectiveValue()
  {
    return objectiveValue;
  }


  /**
   * registers the truth value of a pseudo boolean variable
   */
  void setValue(string variable, bool value)
  {
    if (interpretation.find(variable)!=interpretation.end() &&
	interpretation[variable]!=value)
    {
      cerr << "ERROR: variable " << variable 
	   << " is assigned contradictory values !" << endl;
      exit(1);
    }

    interpretation[variable]=value;
  }

  /**
   * return the truth value of the variable called variable
   * if positive is false, we return the opposite of the variable
   *
   * when the literal is unassigned, we return false in any case
   * unless mustHaveValue is true in which case we throw an exception
   */
  bool getValue(string variable, 
		bool positive,
		bool mustHaveValue=false)
  {
    map<string,bool>::iterator i=interpretation.find(variable);

    if (i==interpretation.end())
    {
      // literal is unassigned
      // cout << "warning: unassigned literal " << name << endl;
      if (mustHaveValue)
	throw runtime_error("literal without value in objective function");

      return false;
    }

    if (positive)
      return (*i).second;
    else
      return !(*i).second;
  }

  /**
   * callback called to indicate the current line number
   */
  virtual void setLineNum(int lineNum)
  {
    line=lineNum;
  }

  /**
   * callback called before we read the objective function
   */
  virtual void beginObjective()
  {
    objectiveValue=0;
  }

  /**
   * callback called before to set if we must minimize or maximize
   * the objective function
   *
   * @param mustBeMinimized: true if the objective function must be
   * minimize, false if it must be maximized
   */
  virtual void objectiveType(bool mustBeMinimized)
  {
    // ignored
  }

  /**
   * callback called when we read a term of the objective function
   *
   * @param coeff: the coefficient of the term
   * @param var: the identifier of the variable
   * @param positiveLiteral: true for a positive literal, false otherwise
   */
  virtual void objectiveTerm(IntegerType coeff, string var, 
			     bool positiveLiteral=true)
  {
    if (getValue(var,positiveLiteral))
      objectiveValue+=coeff;
  }

  /**
   * callback called when we read a constant in the objective function
   *
   * @param coeff: the constant
   */
  virtual void objectiveConstant(IntegerType coeff)
  {
    objectiveValue+=coeff;
  }

  /**
   * callback called when we need to adjust each coefficient of the
   * previous terms (the ones for which constraintTerm was already
   * called). This can happen for example when we have read integer
   * coefficients so far and the parser discovers a 0.3
   * coefficient. To get all coefficient as integers, each coefficient
   * must be multiplied by 10.
   *
   * @param coeff: the integer by which each previous term must be multiplied
   */
  virtual void multPreviousObjectiveTerms(IntegerType coeff)
  {
    objectiveValue*=coeff;
  }

  /**
   * callback called after we've read the objective function
   */
  virtual void endObjective()
  {
    // ignored
  }


  /**
   * callback called before we read a constraint
   */
  virtual void beginConstraint()
  {
    leftTerm=0;
    rightTerm=0;
    constraintName="";
  }

  /**
   * callback called when we read a term of a constraint
   *
   * @param coeff: the coefficient of the term
   * @param var: the identifier of the variable
   * @param positiveLiteral: true for a positive literal, false otherwise
   */
  virtual void constraintTerm(IntegerType coeff, string var,
		      bool positiveLiteral=true)
  {
    if (getValue(var,positiveLiteral))
      leftTerm+=coeff;
  }

  /**
   * callback called when we read a constant on the left side of a constraint
   *
   * @param coeff: the coefficient of the term
   */
  virtual void constraintConstant(IntegerType coeff)
  {
    leftTerm+=coeff;
  }

  /**
   * callback called when we need to adjust each coefficient of the
   * previous terms (the ones for which constraintTerm was already
   * called). This can happen for example when we have read integer
   * coefficients so far and the parser discovers a 0.3
   * coefficient. To get all coefficient as integers, each coefficient
   * must be multiplied by 10.
   *
   * @param coeff: the integer by which each previous term must be multiplied
   */
  virtual void multPreviousConstraintTerms(IntegerType coeff)
  {
    leftTerm*=coeff;
    rightTerm*=coeff;
  }

  /**
   * callback called when we read the relational operator of a constraint
   *
   * @param relop: the relational operator
   */
  virtual void constraintRelOp(string relop)
  {
    relOp=relop;
  }

  /**
   * callback called when we read the right term of a constraint (also
   * known as the degree)
   *
   * @param val: the degree of the constraint
   */
  virtual void constraintRightTerm(IntegerType degree)
  {
    rightTerm=degree;
  }

  /**
   * callback called after we've read a constraint
   */
  virtual void endConstraint()
  {
    bool result=false;

    if (relOp=="=")
      result=leftTerm==rightTerm;
    else
      if (relOp=="<=")
	result=leftTerm<=rightTerm;
      else
	if (relOp=="<")
	  result=leftTerm<rightTerm;
	else
	  if (relOp==">=")
	    result=leftTerm>=rightTerm;
	  else
	    if (relOp==">")
	      result=leftTerm>rightTerm;
	    else
	    {
	      cerr << "INTERNAL ERROR: unexpected relational operator: " 
		   << relOp << endl;
	      exit(127);
	    }

    if (!result)
    {
      cout << "FAILED" << endl;
      cout << "ERROR: unsatisfied constraint on line " << line;
      if (constraintName!="")
	cout << " (contraint name=" << constraintName << ")";
      cout << endl;

      cout << "\t" << leftTerm << relOp << rightTerm << endl;
      exit(6);
    }
#ifdef debug
    else
    {
      cout << "OK\t" << leftTerm << relOp << rightTerm 
	   << " (contraint name=" << constraintName << ")" << endl;
    }
#endif
  }

private:
  ifstream in; // the stream we're reading from
  int nbVars,nbConstr; // MetaData: #Variables and #Constraints in file.
  
  /**
   * get the next character from the stream
   */
  char get()
  {
    return in.get();
  }

  /**
   * put back a character into the stream (only one chr can be put back)
   */
  void putback(char c)
  {
    in.putback(c);
  }

  /**
   * return true iff we've reached EOF
   */
  bool eof()
  {
    return !in.good();
  }

  /**
   * skip white spaces
   */
  void skipSpaces()
  {
    char c;

    while(isspace(c=get()));

    putback(c);
  }

  /**
   * read an identifier from stream and store it in s
   * @param s: the variable to hold the identifier we read
   * @return true iff an identifier was correctly read
   */
  bool readIdentifier(string &s)
  {
    char c;
    int varID;
    
    skipSpaces();

    // first char (must be 'x')
    c=get();
    if (eof())
      return false;

    if (c!='x') {
      putback(c);
      return false;
    }

    s=c;
    
    // next chars (must be digits)
    while(true) {
      c=get();
      if (eof())
	break;
      
      if (isdigit(c))
	s+=c;
      else {
	putback(c);
	break;
      }
    }
    
    //Small check on the coefficient ID to make sure everything is ok
    varID = atoi(s.c_str()+1);
    if (varID > nbVars) {
      throw runtime_error("variable identifier larger than #variables in metadata.");
    }
    
    return true;
  }

  /**
   * read a relational operator from stream and store it in s
   * @param s: the variable to hold the relational operator we read
   * @return true iff a relational operator was correctly read
   */
  bool readRelOp(string &s)
  {
    char c;

    skipSpaces();

    c=get();
    if (eof())
      return false;

    if (c=='=')
    {
      s="=";
      return true;
    }

    if (c=='>' && get()=='=')
    {
      s=">=";
      return true;
    }

    return false;
  }

  /**
   * read the first comment line to get the number of variables and
   * the number of constraints in the file
   *
   * calls metaData with the data that was read
   */
  void readMetaData()
  {
    char c;
    string s;

    // get the number of variables and constraints
    c=get();
    if (c!='*')
      throw runtime_error("First line of input file should be a comment");

    in >> s;
    if (eof() || s!="#variable=")
      throw runtime_error("First line should contain #variable= as first keyword");

    in >> nbVars;

    in >> s;
    if (eof() || s!="#constraint=")
      throw runtime_error("First line should contain #constraint= as second keyword");

    in >> nbConstr;

    // skip the rest of the line
    getline(in,s);
  }


  /**
   * skip the comments at the beginning of the file
   */
  void skipComments()
  {
    string s;
    char c;

    // skip further comments

    while(!eof() && (c=get())=='*')
    {
      getline(in,s);
    }

    putback(c);
  }


  /**
   * read a term into coeff and var
   * @param coeff: the coefficient of the variable
   * @param var: the indentifier we read
   */
  void readTerm(IntegerType &coeff, string &var)
  {
    char c;
    
    in >> coeff;
    
    skipSpaces();
    
    if (!readIdentifier(var))
      throw runtime_error("identifier expected");
  }

  /**
   * read the objective line (if any)
   *
   * calls beginObjective, objectiveTerm and endObjective
   */
  void readObjective()
  {
    char c;
    string s,var;
    IntegerType coeff;

    // read objective line (if any)

    skipSpaces();
    c=get();
    if (c!='m')
    {
      // no objective line
      putback(c);
      return;
    }

    if (get()=='i' && get()=='n' && get()==':')
    {
      beginObjective(); // callback

      while(!eof())
      {
	readTerm(coeff,var);
	objectiveTerm(coeff,var); // callback

	skipSpaces();
	c=get();
	if (c==';')
	  break; // end of objective
	else
	  if (c=='-' || c=='+' || isdigit(c))
	    putback(c);
	  else
	    throw runtime_error("unexpected character in objective function");
      }

      endObjective();
    }
    else
      throw runtime_error("input format error: 'min:' expected");
  }

  /**
   * read a constraint
   *
   * calls beginConstraint, constraintTerm and endConstraint
   */
  void readConstraint()
  {
    string s, var;
    char c;
    IntegerType coeff;

    beginConstraint();

    while(!eof())
    {
      readTerm(coeff,var);
      constraintTerm(coeff,var);

      skipSpaces();
      c=get();
      if (c=='>' || c=='=')
      {
	// relational operator found
	putback(c);
	break;
      }
      else
	if (c=='-' || c=='+' || isdigit(c))
	  putback(c);
	else
	  throw runtime_error("unexpected character in constraint");
    }

    if (eof())
      throw runtime_error("unexpected EOF before end of constraint");
      
    if (readRelOp(s))
      constraintRelOp(s);
    else
      throw runtime_error("unexpected relational operator in constraint");
    
    in >> coeff;
    constraintRightTerm(coeff);
    
    skipSpaces();
    c=get();
    if (eof() || c!=';')
      throw runtime_error("semicolon expected at end of constraint");
    
    endConstraint();
  }

public:
  SimpleChecker()
  {
    objectiveValue=0;
  }


  /**
   * parses the file and uses the callbacks to send to send the data
   * back to the program
   */
  void parse(const char *filename)
  {
    in.open(filename,ios_base::in);

    if (!in.good())
      throw runtime_error("error opening input file");

    readMetaData();

    skipComments();

    readObjective();

    // read constraints
    int nbConstraintsRead = 0;
    while(!eof()) {
      skipSpaces();
      if (eof())
	break;
      
      readConstraint();
      nbConstraintsRead++;
    }
    
    //Small check on the number of constraints
    if (nbConstraintsRead != nbConstr) {
      throw runtime_error("number of constraints read is different from metadata.");
    }
  }
}; // class SimpleChecker




#endif
